"""
transpiler/analyzer.py — Semantic Analysis Pass for IR (v0.5.0).
"""

from __future__ import annotations
import json
import os
from typing import Dict, Optional, Set, Any
from transpiler.middleend import ir
from transpiler.infrastructure.diagnostics import manager, DiagnosticManager
from transpiler.infrastructure.project_context import ProjectContext, ModuleExport

class SymbolTable:
    def __init__(self, parent: Optional[SymbolTable] = None):
        self.parent = parent
        self.symbols: Dict[str, str] = {} # name -> type_name
        self.definitions: Dict[str, tuple[int, int, Optional[str]]] = {} # name -> (line, col, origin_module)

    def lookup(self, name: str) -> Optional[str]:
        if name in self.symbols:
            return self.symbols[name]
        p = self.parent
        if p is not None:
            return p.lookup(name)
        return None

    def lookup_definition(self, name: str) -> Optional[tuple[int, int, Optional[str]]]:
        if name in self.definitions:
            return self.definitions[name]
        p = self.parent
        if p is not None:
            return p.lookup_definition(name)
        return None

    def define(self, name: str, type_name: str, lineno: int = 0, col_offset: int = 0, origin: Optional[str] = None):
        self.symbols[name] = type_name
        self.definitions[name] = (lineno, col_offset, origin)

SDK_MODULES = {"roblox", "math", "table", "string", "task", "coroutine", "debug", "re", "json", "random"}

class SemanticAnalyzer:
    def __init__(self, diagnostics: Optional[DiagnosticManager] = None, project: Optional[ProjectContext] = None, module_name: str = "main", uri: Optional[str] = None):
        self.diagnostics = diagnostics or DiagnosticManager()
        self.project = project
        self.module_name = module_name
        self.uri = uri
        self._class_info: Dict[str, Any] = {}
        self._current_class: Optional[ir.IRClassDef] = None
        
        # Load core snippets if available
        self.metadata = self._load_metadata()
        self.symbol_table = SymbolTable()
        self._init_globals()
        self._exports = ModuleExport(module_name, uri)

    def _load_metadata(self) -> Dict[str, Any]:
        path = os.path.join(os.path.dirname(__file__), "api_metadata.json")
        try:
            with open(path, "r") as f:
                return json.load(f)
        except Exception:
            # Silence metadata loading warnings in LSP mode
            return {"classes": {}, "globals": {}}

    def _init_globals(self):
        globals_map = self.metadata.get("globals", {})
        for name, type_name in globals_map.items():
            self.symbol_table.define(name, type_name)

    def analyze(self, node: ir.IRNode):
        method_name = f"visit_{type(node).__name__}"
        visitor = getattr(self, method_name, self.generic_visit)
        visitor(node)
        
        # Ensure all expressions receive an inferred type after visiting
        if isinstance(node, ir.IRExpression) and getattr(node, "inferred_type", None) is None:
            node.inferred_type = self._infer_type(node)

    def generic_visit(self, node: ir.IRNode):
        if hasattr(node, "__dict__"):
            for attr, value in node.__dict__.items():
                if isinstance(value, ir.IRNode):
                    self.analyze(value)
                elif isinstance(value, list):
                    for item in value:
                        if isinstance(item, ir.IRNode):
                            self.analyze(item)
                elif isinstance(value, ir.IRIfElseExpression):
                    self.analyze(value.condition)
                    self.analyze(value.then_expr)
                    self.analyze(value.else_expr)

    def visit_IRModule(self, node: ir.IRModule):
        # We'll store class property info in class_properties map or similar
        self._class_info = {}

        for stmt in node.body:
            self.analyze(stmt)
        
        # After analysis, if we are in a project, update the global project context
        if self._exports:
            self._exports.symbols = self.symbol_table.symbols.copy()
            self._exports.classes = self._class_info.copy()
            
        proj = self.project
        if proj is not None:
            proj.modules[self.module_name] = self._exports

        # Capture metadata for the generator
        node.metadata = {
            "symbols": self.symbol_table.symbols,
            "classes": self._class_info
        }

    def visit_IRAssignment(self, node: ir.IRAssignment):
        self.analyze(node.value)
        inferred = self._infer_type(node.value)
        if inferred and not node.value.inferred_type:
            node.value.inferred_type = inferred
        
        if isinstance(node.target, ir.IRVariable):
            name = getattr(node.target, "name", None)
            if name:
                t = self._infer_type(node.value) or "any"
                self.symbol_table.define(name, t, node.target.lineno, node.target.col_offset)
                
                # Update inferred_type on target node itself
                node.target.inferred_type = t
                
                # Export if we are at module level
                if self.symbol_table.parent is None: # Toplevel
                    self._exports.symbols[node.target.name] = inferred
                    self._exports.definitions[node.target.name] = (node.target.lineno, node.target.col_offset)
        elif isinstance(node.target, ir.IRPropertyAccess):
            # Check for 'self.prop = value'
            if isinstance(node.target.receiver, ir.IRVariable) and node.target.receiver.name == "self":
                curr_class = getattr(self, "_current_class", None)
                if curr_class:
                    curr_class.properties[node.target.property] = inferred or "any"
                    node.target.inferred_type = inferred

        self.analyze(node.target)
        
        # Type check: if target already has a type, ensure value is compatible
        if isinstance(node.target, ir.IRVariable):
            target_type = self.symbol_table.lookup(node.target.name)
            if target_type and target_type != "any" and inferred and inferred != "any":
                if target_type != inferred:
                    self.diagnostics.error(
                        f"Type mismatch: cannot assign '{inferred}' to variable of type '{target_type}'",
                        node.lineno, node.col_offset, self.uri
                    )

    def visit_IRBinaryOperation(self, node: ir.IRBinaryOperation):
        self.analyze(node.left)
        self.analyze(node.right)
        
        lt = node.left.inferred_type or "any"
        rt = node.right.inferred_type or "any"
        op = node.operator
        
        # Mapping AST names to human-readable symbols for error messages
        OP_DISPLAY = {
            "Add": "+", "Sub": "-", "Mult": "*", "Div": "/", "Mod": "%", "Pow": "**", "FloorDiv": "//",
            "Lt": "<", "LtE": "<=", "Gt": ">", "GtE": ">=", "Eq": "==", "NotEq": "!=",
            "And": "and", "Or": "or"
        }
        display_op = OP_DISPLAY.get(op, op)
        
        # Arithmetic operators: both must be numbers
        if op in ("Add", "Sub", "Mult", "Div", "Mod", "Pow", "FloorDiv"):
            # Attempt to find the operator position (e.g., after the left operand)
            # This is a heuristic: left_end + 1
            op_col = node.col_offset
            if hasattr(node.left, "name"):
                op_col = node.left.col_offset + len(node.left.name) + 1 # +1 for space
            elif isinstance(node.left, ir.IRLiteral):
                op_col = node.left.col_offset + len(str(node.left.value)) + 1
            
            if lt != "any" and lt != "number":
                self.diagnostics.error(f"Operator '{display_op}' is not supported for type '{lt}'", node.lineno, op_col, self.uri)
            if rt != "any" and rt != "number":
                self.diagnostics.error(f"Operator '{display_op}' is not supported for type '{rt}'", node.lineno, op_col, self.uri)
        
        # Comparisons: usually need same types
        elif op in ("Lt", "LtE", "Gt", "GtE", "Eq", "NotEq"):
            if lt != "any" and rt != "any" and lt != rt:
                self.diagnostics.warning(f"Comparison between mismatched types '{lt}' and '{rt}'", node.lineno, node.col_offset, self.uri)

    def _track_usage(self, name: str, node: ir.IRNode):
        # sys.stderr.write(f"DEBUG: _track_usage called for '{name}' in '{self.module_name}'\n")
        if self.project and self.module_name:
            defn = self.symbol_table.lookup_definition(name)
            target_mod = self.module_name
            if defn and defn[2]: # origin is third in the tuple
                target_mod = defn[2]
            
            mod = self.project.get_module(target_mod)
            if name not in mod.usages:
                mod.usages[name] = []
            mod.usages[name].append((self.uri or "unknown", node.lineno, node.col_offset))

    def visit_IRVariable(self, node: ir.IRVariable):
        # Resolve symbol in symbol table
        symbol = self.symbol_table.lookup(node.name)
        if symbol:
            node.inferred_type = symbol
        # Record usage
        self._track_usage(node.name, node)
            
        # Record definition location for LSP
        defn = self.symbol_table.lookup_definition(node.name)
        if defn:
            node.definition_location = defn # (line, col, origin)

    def visit_IRFunctionCall(self, node: ir.IRFunctionCall):
        self.analyze(node.func)
        for arg in node.args:
            self.analyze(arg)
        
        # Handle SDK member calls (e.g., from string import upper -> upper("..."))
        if isinstance(node.func, ir.IRVariable) and node.func.inferred_type:
            if node.func.inferred_type.startswith("SDK_MEMBER:"):
                qualified = node.func.inferred_type[11:]
                if "." in qualified:
                    mod, member = qualified.split(".", 1)
                    # Transform this function call into a method call (qualified library call)
                    # We can't easily replace 'node' itself in a visitor, but we can 
                    # change its 'func' to an Attribute if we had that in IR.
                    # In RPy IR, we have IRPropertyAccess.
                    # Let's change node.func to an IRPropertyAccess.
                    node.func = ir.IRPropertyAccess(
                        receiver=ir.IRVariable(name=mod, lineno=node.lineno, col_offset=node.col_offset),
                        property=member,
                        lineno=node.lineno,
                        col_offset=node.col_offset
                    )

        if not node.inferred_type:
            node.inferred_type = self._infer_type(node)

    def visit_IRMethodCall(self, node: ir.IRMethodCall):
        self.analyze(node.receiver)
        for arg in node.args:
            self.analyze(arg)

        receiver_type = self._infer_type(node.receiver)
        node.call_style = ir.CallStyle.COLON # Default

        # 1. Aggressively map common Roblox patterns regardless of type info
        # as these are almost always intended to be the CamelCase Roblox versions.
        common_rbx_patterns = {
            "connect": "Connect",
            "wait": "Wait",
            "destroy": "Destroy",
            "clone": "Clone",
            "is_a": "IsA",
            "get_service": "GetService",
            "wait_for_child": "WaitForChild",
            "get_children": "GetChildren",
            "find_first_child": "FindFirstChild",
            "set_core_gui_enabled": "SetCoreGuiEnabled"
        }
        if node.method in common_rbx_patterns:
            node.method = common_rbx_patterns[node.method]
        elif "_" in node.method:
            # General snake_case to CamelCase fallback for any colon call
            # This is idiomatic for Python-to-Roblox transpilation.
            node.method = "".join(p.capitalize() for p in node.method.split("_"))

        if receiver_type:
            if receiver_type.startswith("Module:"):
                node.call_style = ir.CallStyle.DOT # Modules use dots
                # Infer return type from module exports
                mod_name = receiver_type[7:]
                proj = self.project
                if proj is not None:
                    mod_export = proj.modules.get(mod_name)
                    if mod_export:
                        t = mod_export.symbols.get(node.method)
                        if t and t.startswith("Function:"):
                            node.inferred_type = t[9:]
                return

            is_optional = receiver_type.endswith("?")
            base_type = receiver_type[:-1] if is_optional else receiver_type
            class_name = base_type[6:] if base_type.startswith("Class:") else base_type

            class_info = self._class_info.get(class_name)
            if class_info:
                # 1. Direct match
                method_type = class_info.get("methods", {}).get(node.method)
                
                # 2. Case-insensitive match for Roblox APIs (if metadata available)
                if not method_type:
                    search_name = node.method.lower()
                    for m_name in class_info.get("methods", {}):
                        if m_name.lower() == search_name:
                            node.method = m_name # Remap to CamelCase
                            method_type = class_info["methods"][m_name]
                            break
                
                if method_type in ("classmethod", "staticmethod"):
                    node.call_style = ir.CallStyle.DOT
                else:
                    node.call_style = ir.CallStyle.COLON
                return

    def visit_IRImport(self, node: ir.IRImport):
        if node.module in SDK_MODULES:
            # e.g., import math -> math is a module
            self.symbol_table.define(node.alias or node.module, f"Module:{node.module}")
            return

        if self.project:
            # We don't perform deep analysis here, just define the module name in symbol table
            self.symbol_table.define(node.alias or node.module, f"Module:{node.module}")

    def visit_IRImportFrom(self, node: ir.IRImportFrom):
        if node.module == "roblox":
            for i, name in enumerate(node.names):
                alias = node.aliases[i] or name
                # e.g., from roblox import task -> task is a module/global
                if name in SDK_MODULES:
                    self.symbol_table.define(alias, f"Module:{name}")
                else:
                    self.symbol_table.define(alias, "any")
            return

        if node.module in SDK_MODULES:
            for i, name in enumerate(node.names):
                alias = node.aliases[i] or name
                # e.g., from math import floor -> floor is a member of math
                self.symbol_table.define(alias, f"SDK_MEMBER:{node.module}.{name}")
            return

        if self.project is not None:
            mod_export = self.project.modules.get(node.module)
            if mod_export:
                for i, name in enumerate(node.names):
                    alias = node.aliases[i] or name
                    t = mod_export.symbols.get(name, "any")
                    self.symbol_table.define(alias, t, origin=node.module)
                    # Track usage of the imported symbol in its source module
                    self._track_usage(alias, node)
            else:
                proj = self.project
                if proj is not None:
                    self.diagnostics.hint(
                        f"Module '{node.module}' not found. (Root: {proj.workspace_root or 'None'})",
                        line=node.lineno, col=node.col_offset
                    )

    def visit_IRPropertyAccess(self, node: ir.IRPropertyAccess):
        self.analyze(node.receiver)
        receiver_type = self._infer_type(node.receiver)
        
        if not receiver_type or receiver_type == "any":
            return

        # --- Module Access Fallback ---
        if receiver_type.startswith("Module:"):
            mod_name = receiver_type[7:]
            proj = self.project
            if proj is not None:
                mod_export = proj.modules.get(mod_name)
                if mod_export:
                    t = mod_export.symbols.get(node.property)
                    if t:
                        node.inferred_type = t
                        return
            
            if mod_name in SDK_MODULES:
                # For SDK modules, we assume the property exists for now
                node.inferred_type = "any"
                return

            self.diagnostics.error(
                f"Module '{mod_name}' has no attribute '{node.property}'.",
                line=node.lineno, col=node.col_offset
            )
            return

        is_optional = receiver_type.endswith("?")
        base_type = receiver_type[:-1] if is_optional else receiver_type
        
        if is_optional:
            receiver_name = getattr(node.receiver, "name", "object") if hasattr(node.receiver, "name") else "object"
            self.diagnostics.warning(
                f"Property access '{node.property}' on optional type '{receiver_type}'.",
                line=node.lineno, col=node.col_offset,
                hint=f"Consider checking 'if {receiver_name} then' first."
            )

        class_info = self._get_class_info(base_type)
        if class_info:
            props = class_info.get("properties", {})
            prop_type = props.get(node.property)
            if prop_type:
                node.inferred_type = prop_type
                return
            
            # --- Service / Child Fallback for Instances ---
            is_instance = base_type in ("Instance", "DataModel")
            if not is_instance:
                # Check inheritance
                curr = base_type
                while curr:
                    info = self._get_class_info(curr)
                    if info and info.get("inherits") in ("Instance", "DataModel"):
                        is_instance = True
                        break
                    curr = info.get("inherits") if info else None

            if is_instance:
                # 1. Try case-insensitive match for properties
                search_name = node.property.replace("_", "").lower()
                for p_name in props:
                    if p_name.lower() == search_name:
                        node.property = p_name
                        node.inferred_type = props[p_name]
                        return
                
                # 2. Assume it's a child or dynamic property
                node.inferred_type = "any"
                return

            # Check signals
            signals = class_info.get("signals", {})
            signal_type = signals.get(node.property)
            if signal_type:
                node.inferred_type = signal_type
                # Track usage of the signal
                self._track_usage(node.property, node)
                return
            
            # Record usage of the property/method
            self._track_usage(node.property, node)
            
            # Not found: Fuzzy match
            from transpiler.frontend.ast_utils import get_closest_match
            all_possibilities = list(props.keys()) + list(signals.keys()) + list(class_info.get("methods", {}).keys())
            closest = get_closest_match(node.property, all_possibilities)
            
            self.diagnostics.error(
                f"Property '{node.property}' does not exist on type '{base_type}'.",
                line=node.lineno, col=node.col_offset,
                hint=f"Did you mean '{closest}'?" if closest else f"Verify the member exists on {base_type}."
            )

    def visit_IRFunctionDef(self, node: ir.IRFunctionDef):
        # Define the function itself in the parent symbol table first
        if node.name:
            # We use a special prefix for function return types in symbols or just store the type
            # For simplicity let's store it as "ReturnType(className)" or similar
            # Actually, let's just use the function name to store its return type for now.
            pass

        old_table = self.symbol_table
        self.symbol_table = SymbolTable(parent=old_table)
        
        # Define arguments with types if available
        for i, arg_name in enumerate(node.args):
            t_ann = node.arg_types[i] if i < len(node.arg_types) else None
            # Arguments are defined at the function start
            # For now we use the function line/col, in a better IR we'd have arg locs
            self.symbol_table.define(arg_name, t_ann or "any", node.lineno, node.col_offset)
        
        for stmt in node.body:
            self.analyze(stmt)
            
        # Optional: Attempt to infer return type if not explicitly set
        if not node.return_type:
            # Simple heuristic: find all return statements
            returns = []
            def find_returns(n):
                if isinstance(n, ir.IRReturnStatement):
                    returns.append(n)
                elif hasattr(n, "__dict__"):
                    for attr, v in n.__dict__.items():
                        if isinstance(v, list):
                            for item in v: find_returns(item)
                        elif isinstance(v, ir.IRNode): find_returns(v)
            
            find_returns(node)
            if returns:
                # Just take the first one for now
                ret_type = self._infer_type(returns[0].value) if returns[0].value else "nil"
                node.return_type = ret_type
            else:
                node.return_type = "any"

        self.symbol_table = old_table
        name = getattr(node, "name", None)
        if name:
            self.symbol_table.define(name, f"Function:{node.return_type}", node.lineno, node.col_offset)
            # Export if we are at module level
            if self.symbol_table.parent is None:
                self._exports.symbols[name] = f"Function:{node.return_type}"
                self._exports.definitions[name] = (node.lineno, node.col_offset)

    def visit_IRGenericForStatement(self, node: ir.IRGenericForStatement):
        self.analyze(node.iterator)
        
        # Heuristic for iterator types
        if not node.var_types:
            it_type = self._infer_type(node.iterator)
            if it_type and it_type.startswith("Class:"):
                class_name = it_type[6:]
                class_info = self._get_class_info(class_name)
                if class_info and "__iter__" in class_info.get("methods", {}):
                    # For now, assume it returns any. In a full impl, we'd check __iter__'s return type.
                    node.var_types = ["any"]
            
            if isinstance(node.iterator, ir.IRFunctionCall) and isinstance(node.iterator.func, ir.IRVariable):
                it_name = node.iterator.func.name
                if it_name == "ipairs":
                    node.var_types = ["number", "any"]
                elif it_name == "pairs":
                    node.var_types = ["any", "any"]
        
        old_table = self.symbol_table
        self.symbol_table = SymbolTable(parent=old_table)
        for i, v in enumerate(node.vars):
            v_type = node.var_types[i] if i < len(node.var_types) else "any"
            self.symbol_table.define(v, v_type, node.lineno, node.col_offset)
        
        for stmt in node.body:
            self.analyze(stmt)
        self.symbol_table = old_table

    def visit_IRWhileStatement(self, node: ir.IRWhileStatement):
        self.analyze(node.condition)
        for stmt in node.body:
            self.analyze(stmt)

    def visit_IRForStatement(self, node: ir.IRForStatement):
        self.analyze(node.start)
        self.analyze(node.end)
        if node.step: self.analyze(node.step)
        
        old_table = self.symbol_table
        self.symbol_table = SymbolTable(parent=old_table)
        self.symbol_table.define(node.var, "number", node.lineno, node.col_offset)
        # S5: Mark loop variable as certain 'number'
        node.var_type_info = ir.TypeMetadata(luau_type="number", confidence="certain")
        
        for stmt in node.body:
            self.analyze(stmt)
        self.symbol_table = old_table

    def _extract_narrowings(self, condition: ir.IRExpression) -> Dict[str, str]:
        """
        Return a dict of {var_name: narrowed_type} that holds in the then-branch
        of an if-statement given `condition`.

        Supported patterns:
        - `x`                     → x: non-optional (strip `?`)
        - `x is None`             → x: "nil"
        - `x is not None`         → x: base type (strip `?`)
        - `isinstance(x, T)`      → x: T  (where T is an IRVariable naming a type)
        - `cond1 and cond2`       → merge both dicts
        - `not cond`              → invert single-var narrowings
        """
        narrowings: Dict[str, str] = {}

        # Pattern 1: bare variable  →  strip optional
        if isinstance(condition, ir.IRVariable):
            name = getattr(condition, "name", None)
            if name:
                existing = self.symbol_table.lookup(name)
                if existing and existing.endswith("?"):
                    narrowings[name] = existing[:-1]
        
        # Pattern 2: `not x`  →  variable is nil/None
        elif isinstance(condition, ir.IRUnaryOperation) and condition.operator in ("not ", "not"):
            if isinstance(condition.operand, ir.IRVariable):
                name = getattr(condition.operand, "name", None)
                if name:
                    existing = self.symbol_table.lookup(name)
                    if existing:
                        narrowings[name] = "nil"

        # Pattern 3: binary comparisons and logical combinators
        elif isinstance(condition, ir.IRBinaryOperation):
            op = condition.operator

            # `x is None` or `x == None`
            if op in ("Is", "Eq", "is", "=="):
                lhs, rhs = condition.left, condition.right
                # normalise order
                if isinstance(rhs, ir.IRVariable) and isinstance(lhs, ir.IRLiteral) and lhs.value is None:
                    lhs, rhs = rhs, lhs
                if isinstance(lhs, ir.IRVariable) and isinstance(rhs, ir.IRLiteral) and rhs.value is None:
                    name = getattr(lhs, "name", None)
                    if name:
                        narrowings[name] = "nil"

            # `x is not None` / `x != None`
            elif op in ("IsNot", "NotEq", "is not", "!="):
                lhs, rhs = condition.left, condition.right
                if isinstance(rhs, ir.IRVariable) and isinstance(lhs, ir.IRLiteral) and lhs.value is None:
                    lhs, rhs = rhs, lhs
                if isinstance(lhs, ir.IRVariable) and isinstance(rhs, ir.IRLiteral) and rhs.value is None:
                    name = getattr(lhs, "name", None)
                    if name:
                        existing = self.symbol_table.lookup(name)
                        if existing and existing.endswith("?"):
                            narrowings[name] = existing[:-1]  # x: T? → T

            # `cond and cond`  →  merge
            elif op in ("And", "and"):
                narrowings.update(self._extract_narrowings(condition.left))
                narrowings.update(self._extract_narrowings(condition.right))

        # Pattern 4: isinstance(x, T)
        elif isinstance(condition, ir.IRFunctionCall):
            if isinstance(condition.func, ir.IRVariable) and getattr(condition.func, "name", None) == "isinstance":
                if len(condition.args) == 2:
                    target = condition.args[0]
                    type_arg = condition.args[1]
                    if isinstance(target, ir.IRVariable) and isinstance(type_arg, ir.IRVariable):
                        # Map Python primitive names to Luau type strings
                        TYPE_MAP = {
                            "int": "number", "float": "number", "str": "string",
                            "bool": "boolean", "list": "{any}", "dict": "{[any]: any}",
                        }
                        t_name = getattr(target, "name", None)
                        ta_name = getattr(type_arg, "name", None)
                        if t_name and ta_name:
                            narrowings[t_name] = TYPE_MAP.get(ta_name, ta_name)

        return narrowings

    def _invert_narrowings(self, then_narrowings: Dict[Any, str]) -> Dict[str, str]:
        """
        Compute else-branch narrowings by inverting then-branch ones.
        If then-branch: x → T      else-branch: x → T? (re-add optional)
        If then-branch: x → nil    else-branch: x → existing base type
        """
        inverted: Dict[str, str] = {}
        for name, narrowed_type in then_narrowings.items():
            if not isinstance(name, str):
                continue
            existing = self.symbol_table.lookup(name)
            if narrowed_type == "nil":
                # then: x is nil  →  else: x is not nil → base type
                if existing and existing.endswith("?"):
                    inverted[name] = existing[:-1]
            elif existing and existing.endswith("?"):
                # then: x is not nil  →  else: x might still be nil
                inverted[name] = existing  # keep optional
            # isinstance narrowings don't invert cleanly → skip (keep outer type)
        return inverted

    def visit_IRIfStatement(self, node: ir.IRIfStatement):
        self.analyze(node.condition)

        # --- Flow-sensitive type narrowing ---
        then_narrowings = self._extract_narrowings(node.condition)
        else_narrowings = self._invert_narrowings(then_narrowings)

        # Collect original types for restoration
        saved_types: Dict[str, str] = {}
        for name in then_narrowings:
            t = self.symbol_table.lookup(name)
            if t is not None:
                saved_types[name] = t

        # Apply then-branch narrowings
        for name, narrowed in then_narrowings.items():
            self.symbol_table.define(name, narrowed)

        for stmt in node.then_block:
            self.analyze(stmt)

        # Restore original types before the else-branch
        for name, orig in saved_types.items():
            self.symbol_table.define(name, orig)

        # Apply else-branch narrowings (if any)
        for name, narrowed in else_narrowings.items():
            self.symbol_table.define(name, narrowed)

        if node.else_block:
            for stmt in node.else_block:
                self.analyze(stmt)

        # Restore after the full if/else block
        for name, orig in saved_types.items():
            self.symbol_table.define(name, orig)


    def visit_IRClassDef(self, node: ir.IRClassDef):
        self.symbol_table.define(node.name, f"Class:{node.name}", node.lineno, node.col_offset)
        
        # Track class info for metadata
        props = []
        meths = {}
        for m in node.body:
            if isinstance(m, ir.IRFunctionDef):
                func_node: ir.IRFunctionDef = m
                if "property" in func_node.decorators:
                    props.append(func_node.name)
                if func_node.name != "__init__":
                    mtype = "method"
                    if "classmethod" in func_node.decorators: mtype = "classmethod"
                    elif "staticmethod" in func_node.decorators: mtype = "staticmethod"
                    meths[func_node.name] = mtype
                else:
                    # Scan __init__ for self.xxx = ...
                    for stmt in func_node.body:
                        if isinstance(stmt, ir.IRAssignment) and isinstance(stmt.target, ir.IRPropertyAccess):
                            if isinstance(stmt.target.receiver, ir.IRVariable) and stmt.target.receiver.name == "self":
                                if stmt.target.property not in props:
                                    props.append(stmt.target.property)

        self._class_info[node.name] = {
            "bases": [str(b) for b in node.bases],
            "properties": props,
            "methods": meths,
            "lineno": node.lineno,
            "col_offset": node.col_offset
        }
        
        if self.symbol_table.parent is None:
            self._exports.symbols[node.name] = f"Class:{node.name}"
            self._exports.definitions[node.name] = (node.lineno, node.col_offset)

        old_class = self._current_class
        self._current_class = node
        
        # We might want a dedicated metadata entry for user classes
        # ... (removed redundancy)
        
        for stmt in node.body:
            self.analyze(stmt)
            
        self._current_class = old_class

    def _infer_type_metadata(self, node: ir.IRExpression) -> Optional[ir.TypeMetadata]:
        if node.type_info: return node.type_info
        
        if isinstance(node, ir.IRLiteral):
            if isinstance(node.value, str): return ir.TypeMetadata("string", "certain")
            if isinstance(node.value, bool): return ir.TypeMetadata("boolean", "certain")
            if isinstance(node.value, (int, float)): return ir.TypeMetadata("number", "certain")
            if node.value is None: return ir.TypeMetadata("nil", "certain")

        if isinstance(node, ir.IRBinaryOperation):
            left_meta = self._infer_type_metadata(node.left)
            right_meta = self._infer_type_metadata(node.right)
            if left_meta and right_meta and left_meta.luau_type == "number" and right_meta.luau_type == "number":
                if node.operator in ("+", "-", "*", "/", "%", "**", "//"):
                    conf = "certain" if (left_meta.confidence == "certain" and right_meta.confidence == "certain") else "inferred"
                    return ir.TypeMetadata("number", conf)
            return ir.TypeMetadata("any", "unknown")

        if isinstance(node, ir.IRIfElseExpression):
            cond_meta = self._infer_type_metadata(node.condition)
            then_meta = self._infer_type_metadata(node.then_expr)
            else_meta = self._infer_type_metadata(node.else_expr)
            
            if then_meta and else_meta and then_meta.luau_type == else_meta.luau_type:
                conf = "certain" if (then_meta.confidence == "certain" and else_meta.confidence == "certain") else "inferred"
                return ir.TypeMetadata(then_meta.luau_type, conf)
            return ir.TypeMetadata("any", "inferred")

        if isinstance(node, ir.IRVariable):
            t = self.symbol_table.lookup(node.name)
            if t: return ir.TypeMetadata(t, "inferred")

        return ir.TypeMetadata("any", "unknown")

    def _infer_type(self, node: ir.IRExpression) -> Optional[str]:
        meta = self._infer_type_metadata(node)
        if meta and meta.luau_type != "any":
            return meta.luau_type
        
        # Analyze specific function/method/property signatures to lift "any"
        if isinstance(node, ir.IRMethodCall):
            if node.method in ("new", "Angles", "fromEulerAnglesYXZ", "fromEulerAnglesXYZ", "fromMatrix", "fromRGB", "fromHSV"):
                if isinstance(node.receiver, ir.IRVariable):
                    constructor_name = node.receiver.name
                    if constructor_name in ("Vector3", "Vector2", "CFrame", "UDim2", "UDim", "Color3", "Rect"):
                        return constructor_name
        elif isinstance(node, ir.IRFunctionCall):
                name = getattr(node.func, "name", None)
                if name:
                    if name in ("Vector3", "Vector2", "CFrame", "UDim2", "UDim", "Color3", "Rect"):
                        return name
                    
                    # Check for class constructor
                    t = self.symbol_table.lookup(name)
                    if t and t.startswith("Class:"):
                        return t
                    
        elif isinstance(node, ir.IRPropertyAccess):
            if node.property in ("Position", "Size", "LookVector", "RightVector", "UpVector"):
                return "Vector3"
            if node.property == "CFrame":
                return "CFrame"
            if node.property == "Color":
                return "Color3"
            if node.property in ("X", "Y", "Z"):
                return "number"

        return meta.luau_type if meta else None

    def _get_class_info(self, class_name: str) -> Optional[Dict[str, Any]]:
        # 1. Check user-defined classes first
        if class_name.startswith("Class:"):
            target = class_name[6:]
            return self._class_info.get(target)

        # 2. Check SDK/Built-in classes
        classes = self.metadata.get("classes", {})
        info = classes.get(class_name)
        if not info:
            return self._class_info.get(class_name) # Fallback
        
        # Merge with base classes if 'inherits' is present
        base_name = info.get("inherits")
        if base_name:
            base_info = self._get_class_info(base_name)
            if base_info:
                # Proper deep merge for properties, methods, signals
                merged = base_info.copy()
                for key in ["properties", "methods", "signals"]:
                    # Handle both list and dict styles
                    if isinstance(info.get(key), list) and isinstance(base_info.get(key), list):
                        merged[key] = list(set(base_info.get(key, []) + info.get(key, [])))
                    else:
                        merged[key] = base_info.get(key, {}).copy()
                        merged[key].update(info.get(key, {}))
                return merged
        return info

def analyze(module: ir.IRModule, diagnostics: Optional[DiagnosticManager] = None, project: Optional[ProjectContext] = None, module_name: str = "main"):
    analyzer = SemanticAnalyzer(diagnostics, project, module_name)
    analyzer.analyze(module)



"""
transpiler/linter.py — Stage 17: Roblox-specific static analysis.
"""

from transpiler.middleend import ir
from transpiler.infrastructure.diagnostics import manager, Severity
from typing import Optional, Set

class LinterPass:
    def __init__(self, diagnostics=None):
        self.diagnostics = diagnostics or manager

    def lint(self, node: ir.IRNode):
        method_name = f"visit_{type(node).__name__}"
        visitor = getattr(self, method_name, self.generic_visit)
        visitor(node)

    def generic_visit(self, node: ir.IRNode):
        for field in node.__dataclass_fields__:
            value = getattr(node, field)
            if isinstance(value, ir.IRNode):
                self.lint(value)
            elif isinstance(value, list):
                for item in value:
                    if isinstance(item, ir.IRNode):
                        self.lint(item)

    def visit_IRFunctionCall(self, node: ir.IRFunctionCall):
        if isinstance(node.func, ir.IRVariable):
            name = node.func.name
            if name == "wait":
                self.diagnostics.warning(
                    "Legacy 'wait()' detected. Use 'task.wait()' for better performance and reliability.",
                    line=node.lineno, col=node.col_offset,
                    hint="task.wait() respects the 60Hz heartbeat and prevents throttling issues."
                )
            elif name == "spawn":
                self.diagnostics.warning(
                    "Legacy 'spawn()' detected. Use 'task.spawn()' for predictable execution order.",
                    line=node.lineno, col=node.col_offset,
                    hint="task.spawn() executes the thread immediately without the 0.03s delay."
                )
            elif name == "delay":
                self.diagnostics.warning(
                    "Legacy 'delay()' detected. Use 'task.delay()'.",
                    line=node.lineno, col=node.col_offset
                )
        self.generic_visit(node)

    def visit_IRVariable(self, node: ir.IRVariable):
        if node.name == "_G":
            self.diagnostics.warning(
                "Usage of global table '_G' detected. This is generally discouraged in Roblox.",
                line=node.lineno, col=node.col_offset,
                hint="Consider using ModuleScripts for shared state management."
            )
        elif node.name == "shared":
             self.diagnostics.warning(
                "Usage of 'shared' table detected. This is generally discouraged in Roblox.",
                line=node.lineno, col=node.col_offset,
                hint="Consider using ModuleScripts for shared state management."
            )

    def visit_IRModule(self, node: ir.IRModule):
        self.generic_visit(node)

    def visit_IRFunctionDef(self, node: ir.IRFunctionDef):
        self.generic_visit(node)

    def visit_IRIfStatement(self, node: ir.IRIfStatement):
        self._check_truthiness(node.condition)
        self.generic_visit(node)

    def visit_IRWhileStatement(self, node: ir.IRWhileStatement):
        self._check_truthiness(node.condition)
        self.generic_visit(node)

    def _check_truthiness(self, node: ir.IRExpression):
        # Python literal 0 and empty string "" are truthy in Lua but falsy in Python.
        # RPy wraps these in py_bool() usually, but if the user uses them directly
        # in a way that bypasses our safety checks (rare but possible in IR), we warn.
        if isinstance(node, ir.IRLiteral):
            if node.value == 0:
                self.diagnostics.warning(
                    "Literal '0' detected in truthiness context. In Lua, 0 is truthy.",
                    line=node.lineno, col=node.col_offset,
                    hint="RPy usually handles this, but explicit 'x != 0' is safer for AAA scale code."
                )
            elif node.value == "":
                self.diagnostics.warning(
                    "Literal empty string detected in truthiness context. In Lua, '' is truthy.",
                    line=node.lineno, col=node.col_offset,
                    hint="Use 'len(s) > 0' for unambiguous cross-language intent."
                )

    def visit_IRIndexOperation(self, node: ir.IRIndexOperation):
        # Off-by-one detection: Python range starts at 0, Lua at 1.
        # If we see a literal 0 being used to index, it's a high-probability bug
        # since RPy doesn't (yet) automatically offset every single index access.
        if isinstance(node.index, ir.IRLiteral) and node.index.value == 0:
            self.diagnostics.warning(
                "Index 0 detected. Roblox Luau uses 1-based indexing.",
                line=node.lineno, col=node.col_offset,
                hint="Python's s[0] becomes s[1] in Lua. Ensure this index is intended."
            )
        self.generic_visit(node)

    def visit_IRForStatement(self, node: ir.IRForStatement):
        self.generic_visit(node)

    def visit_IRGenericForStatement(self, node: ir.IRGenericForStatement):
        self.generic_visit(node)

def lint_ir(module: ir.IRModule, diagnostics=None):
    linter = LinterPass(diagnostics=diagnostics)
    linter.lint(module)



"""
RPy — Python-to-Roblox Luau transpiler.
"""
__version__ = "1.0.0"

from transpiler.frontend import node_registry
from transpiler.frontend import handlers
from transpiler.infrastructure import errors
from transpiler.infrastructure import flags

# ── Backwards-compatible re-exports ──────────────────────────────
# After refactoring the transpiler into sub-packages (frontend, middleend,
# backend, optimization, infrastructure), many existing files still import
# from the old flat paths like `transpiler.generator` or `transpiler.flags`.
# The shim modules below keep those paths working:
#
#   Old path                    → Re-exported from
#   transpiler.ir               → transpiler.middleend.ir
#   transpiler.generator        → transpiler.backend.generator
#   transpiler.flags            → transpiler.infrastructure.flags
#   transpiler.errors           → transpiler.infrastructure.errors
#   transpiler.parser           → transpiler.frontend.parser
#   transpiler.ast_utils        → transpiler.frontend.ast_utils
#   transpiler.node_registry    → transpiler.frontend.node_registry
#   transpiler.ir_optimizer     → transpiler.optimization.ir_optimizer
#   transpiler.query_service    → transpiler.middleend.query_service
#
# These are provided for backwards compatibility only. New code should use
# the fully-qualified sub-package paths.

import importlib as _importlib
import sys as _sys

_COMPAT_MAP = {
    "transpiler.ir":             "transpiler.middleend.ir",
    "transpiler.generator":      "transpiler.backend.generator",
    "transpiler.parser":         "transpiler.frontend.parser",
    "transpiler.ast_utils":      "transpiler.frontend.ast_utils",
    "transpiler.ir_optimizer":   "transpiler.optimization.ir_optimizer",
    "transpiler.query_service":  "transpiler.middleend.query_service",
}

for _alias, _real in _COMPAT_MAP.items():
    if _alias not in _sys.modules:
        try:
            _sys.modules[_alias] = _importlib.import_module(_real)
        except ImportError:
            pass  # Don't crash if a sub-module has its own issues

# `transpiler.flags` and `transpiler.errors` are already importable because
# we import them in this file above together with the sub-packages. But the
# modules themselves are registered under the sub-package name.  We alias
# them so that `from transpiler.flags import CompilerFlags` works:
_sys.modules.setdefault("transpiler.flags", flags)
_sys.modules.setdefault("transpiler.errors", errors)



"""
transpiler/loop_detector.py — Natural Loop Detection for CFG.
"""

from __future__ import annotations
from typing import Dict, Set, List, Optional, TYPE_CHECKING
from transpiler.middleend import cfg
if TYPE_CHECKING:
    from transpiler.middleend import ssa

class LoopInfo:
    def __init__(self, header_id: int):
        self.header_id = header_id
        self.body_ids: Set[int] = set()
        self.parent: Optional[LoopInfo] = None
        self.children: List[LoopInfo] = []

    def __repr__(self):
        return f"Loop(header={self.header_id}, body={sorted(list(self.body_ids))})"

class LoopDetector:
    def __init__(self, target_cfg: cfg.CFG, analysis: ssa.DominatorAnalysis):
        self.cfg = target_cfg
        self.analysis = analysis
        self.loops: Dict[int, LoopInfo] = {} # header_id -> LoopInfo

    def run(self) -> Dict[int, LoopInfo]:
        """Identifies natural loops in the CFG and returns a map of header_id -> LoopInfo."""
        # 1. Find back-edges and identify loops
        # A back-edge is an edge (u -> v) where v dominates u.
        for bid, block in self.cfg.blocks.items():
            for succ in block.successors:
                if succ.id in self.analysis.dominators.get(bid, set()):
                    self._add_to_loop(succ.id, bid)

        # 2. Reconstruct nesting structure
        self._nest_loops()
        return self.loops

    def _add_to_loop(self, header_id: int, back_edge_node_id: int):
        if header_id not in self.loops:
            self.loops[header_id] = LoopInfo(header_id)
        
        loop = self.loops[header_id]
        loop.body_ids.add(header_id)
        
        # Traverse backwards from the back-edge node to finding all nodes in the loop
        stack = [back_edge_node_id]
        while stack:
            n_id = stack.pop()
            if n_id not in loop.body_ids:
                loop.body_ids.add(n_id)
                block = self.cfg.blocks.get(n_id)
                if block:
                    for pred in block.predecessors:
                        stack.append(pred.id)

    def _nest_loops(self):
        """Builds a parent-child relationship tree for nested loops."""
        # Sort loops by body size (inner loops have smaller bodies)
        sorted_loops = sorted(self.loops.values(), key=lambda l: len(l.body_ids))
        for i, inner in enumerate(sorted_loops):
            for outer in sorted_loops[i+1:]:
                # Check if it's a strict subset to avoid self-nesting if header matches (rare)
                if inner.body_ids.issubset(outer.body_ids) and inner.header_id != outer.header_id:
                    inner.parent = outer
                    outer.children.append(inner)
                    break

def detect_loops(target_cfg: cfg.CFG, analysis: "ssa.DominatorAnalysis") -> Dict[int, LoopInfo]:
    detector = LoopDetector(target_cfg, analysis)
    return detector.run()



"""
sdk/table.py — Luau table library stubs for RPy.
"""

from typing import Any, Optional, Callable

def insert(t: list, pos_or_value: Any, value: Any = ...) -> None: ...
def remove(t: list, pos: int = ...) -> Any: ...
def sort(t: list, comp: Callable[[Any, Any], bool] = ...) -> None: ...
def concat(t: list, sep: str = "", i: int = 1, j: int = ...) -> str: ...
def pack(*args: Any) -> list: ...
def unpack(t: list, i: int = 1, j: int = ...) -> Any: ...
def find(t: list, value: Any, init: int = 1) -> Optional[int]: ...
def create(count: int, value: Any = ...) -> list: ...
def clear(t: dict | list) -> None: ...
def move(a1: list, f: int, e: int, t: int, a2: list = ...) -> list: ...
def foreachi(t: list, f: Callable[[int, Any], None]) -> None: ...
def foreach(t: dict, f: Callable[[Any, Any], None]) -> None: ...
def getn(t: list) -> int: ...



import asyncio
import time
from pathlib import Path
from typing import Optional

from cli.main import transpile_file, _get_script_type, _validate_placement, transpile_and_write
from transpiler.flags import CompilerFlags
from transpiler.errors import TranspileError
from sync.server import DevServer
from sync.watcher import start_watcher

class RPyLiveCoordinator:
    def __init__(self, src_dir: Path, out_dir: Path, flags: CompilerFlags, host="localhost", port=8000):
        self.src_dir = src_dir
        self.out_dir = out_dir
        self.flags = flags
        
        # Load config
        project_dir = src_dir if src_dir.is_dir() else src_dir.parent
        from cli.main import _load_rpy_config
        self.config = _load_rpy_config(project_dir)
        
        # Merge CLI flags into config for the plugin
        self.config["flags"] = self.config.get("flags", {})
        self.config["flags"]["backup_studio"] = flags.backup_studio
        
        self.server = DevServer(host, port, config=self.config)
        self.watcher = None

    async def run(self):
        # 1. Initial full build
        print(f"Performing initial build of {self.src_dir}...")
        self.rebuild_all()
        
        # 2. Start server
        await self.server.start()
        
        # 3. Start watcher
        self.watcher = start_watcher(self.src_dir, self.handle_file_change)
        
        print(f"Live sync is active. Press Ctrl+C to stop.")
        
        try:
            while True:
                await asyncio.sleep(1)
        except asyncio.CancelledError:
            self.watcher.stop()
            self.watcher.join()

    def handle_file_change(self, py_path: Path, deleted: bool = False):
        if deleted:
            rel_path = str(py_path.relative_to(self.src_dir))
            if self.server.remove_file(rel_path):
                print(f"  - {rel_path} deleted")
            return

        from transpiler.infrastructure.dependency_graph import DependencyGraph
        from cli.main import _get_project_folders
        valid_roots = list(_get_project_folders(self.server.config).keys())
        graph = DependencyGraph(str(self.src_dir), valid_roots=valid_roots)
        try:
            graph.scan_project()
        except TranspileError as e:
            print(f"  ⚠ DAG validation error: {e}")

        visited = set()
        to_sync = [py_path]
        
        while to_sync:
            curr_path = to_sync.pop(0)
            if curr_path in visited: continue
            visited.add(curr_path)
            
            # Add dependents
            for dep in graph.reverse_graph.get(curr_path, []):
                if dep not in visited and dep not in to_sync:
                    to_sync.append(dep)
                    
            self._sync_single_file(curr_path, is_dependency=(curr_path != py_path))

    def _sync_single_file(self, py_path: Path, is_dependency: bool = False):
        rel_path = str(py_path.relative_to(self.src_dir))
        warning = _validate_placement(py_path, self.src_dir, {})
        if warning and not is_dependency:
            print(f"  ⚠ {warning}")

        start_time = time.time()
        try:
            from cli.main import transpile_file
            # Note: transpile_file expects flags, we pass self.flags
            result = transpile_file(py_path, self.flags)
            latency = (time.time() - start_time) * 1000 # ms
            
            if self.flags.show_out:
                lua_path = self.out_dir / py_path.relative_to(self.src_dir).with_suffix(".lua")
                lua_path.parent.mkdir(parents=True, exist_ok=True)
                lua_path.write_text(result.code, encoding="utf-8")

            script_type = _get_script_type(py_path)
            self.server.update_file(rel_path, result.code, script_type, latency)
            
            prefix = "  ↻" if is_dependency else "  ✓"
            print(f"{prefix} {rel_path} synced ({latency:.1f}ms)")
            
        except TranspileError as e:
            print(f"  ✗ {py_path.name}: {e}")
        except Exception as e:
            print(f"  ✗ {py_path.name}: Internal error: {e}")

    def _sync_runtime(self):
        """Sync the core RPy runtime to the dev server's VFS."""
        runtime_src = Path(__file__).parent.parent / "runtime" / "python_runtime.lua"
        if not runtime_src.exists():
            # Try alternative location if installed as a package
            import rpy
            runtime_src = Path(rpy.__file__).parent / "runtime" / "python_runtime.lua"
            
        if runtime_src.exists():
            code = runtime_src.read_text(encoding="utf-8")
            # Sync to all service folders to ensure accessibility
            for folder in ["replicatedstorage", "serverscriptservice", "starterplayerscripts", "workspace"]:
                path = f"{folder}/python_runtime.lua"
                self.server.update_file(path, code, "module", 0)
            print("  ✓ RPy Runtime synced to VFS")

    def rebuild_all(self):
        self._sync_runtime()
        py_files = []
        # Filter directories to avoid venv, .git, etc.
        for py_file in self.src_dir.rglob("*.py"):
            path_str = str(py_file.absolute())
            if "__pycache__" in path_str or "venv" in path_str or ".git" in path_str or ".rpy" in path_str:
                continue
            if py_file.name == "__init__.py":
                continue
            py_files.append(py_file)
            
        for py_file in py_files:
            self._sync_single_file(py_file)

async def start_live_sync(src: Path, out: Path, flags: CompilerFlags):
    coordinator = RPyLiveCoordinator(src, out, flags)
    await coordinator.run()



"""
transpiler/purity.py — Purity Analysis and Effect Lattice.
"""

from transpiler.middleend import ir
from typing import Dict, Set

class PurityAnalyzer:
    def __init__(self):
        # Whitelist of pure functions / built-ins
        self.pure_builtins = {
            "math.abs", "math.acos", "math.asin", "math.atan", "math.atan2",
            "math.ceil", "math.cos", "math.cosh", "math.deg", "math.exp",
            "math.floor", "math.fmod", "math.frexp", "math.ldexp", "math.log",
            "math.log10", "math.max", "math.min", "math.modf", "math.pow",
            "math.rad", "math.sin", "math.sinh", "math.sqrt", "math.tan",
            "math.tanh", "type", "typeof", "len", "py_len", "str", "py_str",
            "int", "py_int", "float", "py_float"
        }
        
    def analyze_module(self, module: ir.IRModule):
        """Annotates all expressions in the module with their effect level."""
        for stmt in module.body:
            self.analyze_stmt(stmt)
            
    def analyze_stmt(self, stmt: ir.IRStatement):
        if isinstance(stmt, ir.IRAssignment):
            self.analyze_expr(stmt.target)
            self.analyze_expr(stmt.value)
        elif isinstance(stmt, ir.IRIfStatement):
            self.analyze_expr(stmt.condition)
            for s in stmt.then_block: self.analyze_stmt(s)
            for cond, body in stmt.elif_blocks:
                self.analyze_expr(cond)
                for s in body: self.analyze_stmt(s)
            if stmt.else_block:
                for s in stmt.else_block: self.analyze_stmt(s)
        elif isinstance(stmt, (ir.IRWhileStatement, ir.IRForStatement)):
            if hasattr(stmt, "condition"): self.analyze_expr(stmt.condition)
            if hasattr(stmt, "start"): self.analyze_expr(stmt.start)
            if hasattr(stmt, "end"): self.analyze_expr(stmt.end)
            if hasattr(stmt, "step") and stmt.step: self.analyze_expr(stmt.step)
            for s in stmt.body: self.analyze_stmt(s)
        elif isinstance(stmt, ir.IRFunctionDef):
            # We don't do interprocedural purity yet, assume unknown/has side effects
            # but we still analyze the body to annotate internal expressions
            for s in stmt.body: self.analyze_stmt(s)
        elif isinstance(stmt, ir.IRReturnStatement) and stmt.value:
            self.analyze_expr(stmt.value)
        elif isinstance(stmt, ir.IRModule):
            for s in stmt.body: self.analyze_stmt(s)
        elif hasattr(stmt, "__dict__"):
            # Generic traversal for other statement types (Break, Continue, etc)
            pass

    def analyze_expr(self, expr: ir.IRExpression) -> ir.EffectLevel:
        if isinstance(expr, ir.IRLiteral):
            expr.effect_level = ir.EffectLevel.PURE
        elif isinstance(expr, ir.IRVariable):
            # Reading a variable is pure (SSA handles definitions/uses)
            expr.effect_level = ir.EffectLevel.PURE
        elif isinstance(expr, ir.IRBinaryOperation):
            l = self.analyze_expr(expr.left)
            r = self.analyze_expr(expr.right)
            expr.effect_level = self._combine_effects(l, r)
        elif isinstance(expr, ir.IRUnaryOperation):
            o = self.analyze_expr(expr.operand)
            expr.effect_level = o
        elif isinstance(expr, ir.IRFunctionCall):
            # 1. Analyze arguments
            max_arg_effect = ir.EffectLevel.PURE
            for arg in expr.args:
                max_arg_effect = self._combine_effects(max_arg_effect, self.analyze_expr(arg))
            
            # 2. Check function purity
            fn_name = self._get_fn_name(expr.func)
            if fn_name in self.pure_builtins:
                # Built-in pure functions are PURE if their arguments are PURE.
                # If args READ_MEMORY, then the call READS_MEMORY.
                expr.effect_level = self._combine_effects(ir.EffectLevel.PURE, max_arg_effect)
            else:
                # Unknown calls are assumed to have side effects
                expr.effect_level = ir.EffectLevel.HAS_SIDE_EFFECT
                
        elif isinstance(expr, ir.IRMethodCall):
            # Recurse
            self.analyze_expr(expr.receiver)
            for arg in expr.args: self.analyze_expr(arg)
            # Methods can trigger metamethods or mutate state
            expr.effect_level = ir.EffectLevel.HAS_SIDE_EFFECT
            
        elif isinstance(expr, ir.IRPropertyAccess):
            self.analyze_expr(expr.receiver)
            # Property access can trigger __index metamethod
            expr.effect_level = ir.EffectLevel.READS_MEMORY
            
        elif isinstance(expr, ir.IRIndexOperation):
            self.analyze_expr(expr.receiver)
            self.analyze_expr(expr.index)
            # Indexing can trigger __index
            expr.effect_level = ir.EffectLevel.READS_MEMORY
            
        elif isinstance(expr, ir.IRTableLiteral):
            max_eff = ir.EffectLevel.PURE
            for el in expr.elements:
                max_eff = self._combine_effects(max_eff, self.analyze_expr(el))
            for k, v in expr.fields:
                max_eff = self._combine_effects(max_eff, self.analyze_expr(v))
            # Table allocation is deterministic but introduces new identity.
            # For LICM/CSE we can treat it as PURE-ish if elements are pure,
            # but to be safe with identity we'll call it PURE + allocation overhead.
            expr.effect_level = max_eff
            
        elif isinstance(expr, ir.IRPhi):
            max_phi_eff = ir.EffectLevel.PURE
            for v in expr.options.values():
                max_phi_eff = self._combine_effects(max_phi_eff, self.analyze_expr(v))
            expr.effect_level = max_phi_eff
            
        elif isinstance(expr, ir.IRIfElseExpression):
            cond_eff = self.analyze_expr(expr.condition)
            then_eff = self.analyze_expr(expr.then_expr)
            else_eff = self.analyze_expr(expr.else_expr)
            expr.effect_level = self._combine_effects(cond_eff, self._combine_effects(then_eff, else_eff))
            
        else:
            # Fallback for complex/unknown expressions: gather effects from all children
            max_fallback_eff = ir.EffectLevel.PURE
            if hasattr(expr, "get_children"):
                for child in expr.get_children():
                    if isinstance(child, ir.IRExpression):
                        max_fallback_eff = self._combine_effects(max_fallback_eff, self.analyze_expr(child))
            expr.effect_level = self._combine_effects(max_fallback_eff, ir.EffectLevel.HAS_SIDE_EFFECT)
            
        return expr.effect_level

    def _combine_effects(self, e1: ir.EffectLevel, e2: ir.EffectLevel) -> ir.EffectLevel:
        # Lattice: PURE < READS < WRITES < HAS_SIDE_EFFECT < UNKNOWN
        return ir.EffectLevel(max(e1.value, e2.value))

    def _get_fn_name(self, expr: ir.IRExpression) -> str:
        if isinstance(expr, ir.IRVariable):
            return expr.name
        if isinstance(expr, ir.IRPropertyAccess):
            recv = self._get_fn_name(expr.receiver)
            if recv:
                return f"{recv}.{expr.property}"
        return ""

def analyze_purity(module: ir.IRModule):
    analyzer = PurityAnalyzer()
    analyzer.analyze_module(module)



"""
transpiler/handlers.py — Optional AST node handlers registered via decorators.

This module is imported by transpiler.node_registry to populate the central
dispatch maps.  It allows us to split the massive generator.py into smaller,
feature-specific files.

Currently, Phase 13 features (Sets, Match/Case) are still in LuauGenerator,
but core infrastructure is moving here.
"""

from __future__ import annotations
import ast
from typing import TYPE_CHECKING
from transpiler.frontend.node_registry import statement, expression

if TYPE_CHECKING:
    from transpiler.backend.generator import GeneratorContext

# ---------------------------------------------------------------------------
# Core Statement Handlers
# ---------------------------------------------------------------------------

@statement(ast.Pass)
def handle_pass(node: ast.Pass, ctx: GeneratorContext) -> None:
    ctx.emitter.line("-- pass")

@statement(ast.Break)
def handle_break(node: ast.Break, ctx: GeneratorContext) -> None:
    flag = ctx.get_current_break_flag()
    if flag:
        ctx.emitter.line(f"{flag} = true")
    ctx.emitter.line("break")

@statement(ast.Continue)
def handle_continue(node: ast.Continue, ctx: GeneratorContext) -> None:
    ctx.emitter.line("continue")



"""
transpiler/ir.py — Intermediate Representation (IR) for RPy.
"""

from __future__ import annotations
from dataclasses import dataclass, field
from typing import List, Optional, Union, Dict, Any
from enum import Enum, auto

class CallStyle(Enum):
    DOT = auto()
    COLON = auto()
    UNKNOWN = auto()

class EffectLevel(Enum):
    PURE = auto()
    READS_MEMORY = auto()
    WRITES_MEMORY = auto()
    HAS_SIDE_EFFECT = auto()
    UNKNOWN = auto()

@dataclass(kw_only=True)
class IRNode:
    """Base class for all IR nodes."""
    lineno: int = 0
    col_offset: int = 0

@dataclass
class TypeMetadata:
    luau_type: str  # 'number', 'string', 'boolean', 'any', 'nil'
    confidence: str # 'certain' | 'inferred' | 'unknown'

# --- Expressions ---

@dataclass(kw_only=True)
class IRExpression(IRNode):
    inferred_type: Optional[str] = None
    type_info: Optional[TypeMetadata] = None
    effect_level: EffectLevel = EffectLevel.UNKNOWN

@dataclass(kw_only=True)
class IRLiteral(IRExpression):
    value: Any = None
    effect_level: EffectLevel = EffectLevel.PURE
    
    def __repr__(self):
        return f"Lit({repr(self.value)})"

@dataclass(kw_only=True)
class IRVariable(IRExpression):
    name: str = ""
    original_name: Optional[str] = None # Used for SSA renaming
    definition_location: Optional[tuple[int, int, Optional[str]]] = None # (line, col, origin_module)
    effect_level: EffectLevel = EffectLevel.PURE
    
    def __repr__(self):
        return f"Var({self.name})"

@dataclass(kw_only=True)
class IRBinaryOperation(IRExpression):
    left: IRExpression = field(default_factory=lambda: IRLiteral(value=0))
    operator: str = "+"
    right: IRExpression = field(default_factory=lambda: IRLiteral(value=0))
    
    def __repr__(self):
        return f"({self.left} {self.operator} {self.right})"

@dataclass(kw_only=True)
class IRUnaryOperation(IRExpression):
    operator: str = "-"
    operand: IRExpression = field(default_factory=lambda: IRLiteral(value=0))
    
    def __repr__(self):
        return f"{self.operator}({self.operand})"

@dataclass(kw_only=True)
class IRFunctionCall(IRExpression):
    func: IRExpression = field(default_factory=lambda: IRVariable(name=""))
    args: List[IRExpression] = field(default_factory=list)
    
    def __repr__(self):
        return f"{self.func}({', '.join(map(str, self.args))})"

@dataclass(kw_only=True)
class IRMethodCall(IRExpression):
    receiver: IRExpression = field(default_factory=lambda: IRVariable())
    method: str = ""
    args: List[IRExpression] = field(default_factory=list)
    call_style: CallStyle = CallStyle.UNKNOWN
    
    def __repr__(self):
        sep = ":" if self.call_style == CallStyle.COLON else "."
        return f"{self.receiver}{sep}{self.method}({', '.join(map(str, self.args))})"

@dataclass(kw_only=True)
class IRSuperCall(IRExpression):
    method: str = ""
    args: List[IRExpression] = field(default_factory=list)
    
    def __repr__(self):
        return f"super().{self.method}({', '.join(map(str, self.args))})"

@dataclass(kw_only=True)
class IRPropertyAccess(IRExpression):
    receiver: IRExpression = field(default_factory=lambda: IRVariable())
    property: str = ""
    
    def __repr__(self):
        return f"{self.receiver}.{self.property}"

@dataclass(kw_only=True)
class IRTableLiteral(IRExpression):
    elements: List[IRExpression] = field(default_factory=list)
    fields: List[tuple[str, IRExpression]] = field(default_factory=list)
    
    def __repr__(self):
        parts = [str(e) for e in self.elements] + [f"{k}={v}" for k, v in self.fields]
        return "{" + ", ".join(parts) + "}"

@dataclass(kw_only=True)
class IRIndexOperation(IRExpression):
    receiver: IRExpression = field(default_factory=lambda: IRVariable())
    index: IRExpression = field(default_factory=lambda: IRLiteral(value=0))
    
    def __repr__(self):
        return f"{self.receiver}[{self.index}]"

@dataclass(kw_only=True)
class IRPhi(IRExpression):
    # block_id -> versioned variable or expression
    options: Dict[int, IRExpression] = field(default_factory=dict)
    
    def __repr__(self):
        val_str = ", ".join(f"B{bid}:{opt}" for bid, opt in self.options.items())
        return f"Phi({val_str})"

@dataclass(kw_only=True)
class IRIfElseExpression(IRExpression):
    condition: IRExpression = field(default_factory=lambda: IRLiteral(value=True))
    then_expr: IRExpression = field(default_factory=lambda: IRLiteral())
    else_expr: IRExpression = field(default_factory=lambda: IRLiteral())

    def __repr__(self):
        return f"({self.then_expr} if {self.condition} else {self.else_expr})"

@dataclass(kw_only=True)
class IRAwait(IRExpression):
    value: IRExpression = field(default_factory=lambda: IRVariable())
    
    def __repr__(self):
        return f"await {self.value}"

# --- Statements ---

@dataclass(kw_only=True)
class IRStatement(IRNode):
    pass

@dataclass(kw_only=True)
class IRModule(IRStatement):
    body: List[IRStatement] = field(default_factory=list)
    metadata: Dict[str, Any] = field(default_factory=dict)

@dataclass(kw_only=True)
class IRAssignment(IRStatement):
    target: IRExpression = field(default_factory=lambda: IRVariable())
    value: IRExpression = field(default_factory=lambda: IRLiteral())
    annotation: Optional[str] = None
    is_local: bool = False

    
    def __repr__(self):
        return f"{self.target} = {self.value}"

@dataclass(kw_only=True)
class IRPersistentAssign(IRStatement):
    target: IRExpression = field(default_factory=lambda: IRVariable())
    value: IRExpression = field(default_factory=lambda: IRLiteral())
    
    def __repr__(self):
        return f"PERSISTENT {self.target} = {self.value}"

@dataclass(kw_only=True)
class IRFunctionDef(IRStatement):
    name: str = ""
    args: List[str] = field(default_factory=list)
    body: List[IRStatement] = field(default_factory=list)
    is_local: bool = True
    is_method: bool = False
    arg_types: List[Optional[str]] = field(default_factory=list)
    defaults: List[Optional[IRExpression]] = field(default_factory=list)
    return_type: Optional[str] = None
    decorators: List[str] = field(default_factory=list)
    
    def __repr__(self):
        return f"def {self.name}({', '.join(self.args)})"

@dataclass(kw_only=True)
class IRAsyncFunctionDef(IRFunctionDef):
    def __repr__(self):
        return f"async def {self.name}({', '.join(self.args)})"

@dataclass(kw_only=True)
class IRLambda(IRExpression):
    args: List[str] = field(default_factory=list)
    body: List[IRStatement] = field(default_factory=list)
    arg_types: List[Optional[str]] = field(default_factory=list)
    defaults: List[IRExpression] = field(default_factory=list)
    return_type: Optional[str] = None

@dataclass(kw_only=True)
class IRIfStatement(IRStatement):
    condition: IRExpression = field(default_factory=lambda: IRLiteral(value=True))
    then_block: List[IRStatement] = field(default_factory=list)
    elif_blocks: List[tuple[IRExpression, List[IRStatement]]] = field(default_factory=list)
    else_block: Optional[List[IRStatement]] = None
    
    def __repr__(self):
        return f"if {self.condition}"

@dataclass(kw_only=True)
class IRWhileStatement(IRStatement):
    condition: IRExpression = field(default_factory=lambda: IRLiteral(value=True))
    body: List[IRStatement] = field(default_factory=list)
    
    def __repr__(self):
        return f"while {self.condition}"

@dataclass(kw_only=True)
class IRForStatement(IRStatement):
    var: str = ""
    start: IRExpression = field(default_factory=lambda: IRLiteral(value=0))
    end: IRExpression = field(default_factory=lambda: IRLiteral(value=10))
    body: List[IRStatement] = field(default_factory=list)
    step: Optional[IRExpression] = None
    
    def __repr__(self):
        body_str = "\n".join(f"      {s}" for s in self.body)
        return f"for {self.var} in {self.start}..{self.end}:\n{body_str}"

@dataclass(kw_only=True)
class IRGenericForStatement(IRStatement):
    vars: List[str] = field(default_factory=list)
    var_types: List[Optional[str]] = field(default_factory=list)
    iterator: IRExpression = field(default_factory=lambda: IRVariable(name=""))
    body: List[IRStatement] = field(default_factory=list)

@dataclass(kw_only=True)
class IRReturnStatement(IRStatement):
    value: Optional[IRExpression] = None
    
    def __repr__(self):
        return f"return {self.value}"

@dataclass(kw_only=True)
class IRMultiReturn(IRStatement):
    """return a, b, c  — native Luau multi-return.
    Emitted when Python 'return <tuple>' is detected.
    """
    values: List[IRExpression] = field(default_factory=list)

    def __repr__(self):
        return f"return {', '.join(map(str, self.values))}"


@dataclass(kw_only=True)
class IRMultiAssign(IRStatement):
    """local x, y = f()  OR  local x, y = a, b

    Shapes:
      rhs_call is set   → local <targets> = <call>      (unpack multi-return)
      rhs_values        → local <targets> = <v1>, <v2>  (parallel assign)
    """
    targets: List[IRExpression] = field(default_factory=list)
    rhs_call: Optional[IRExpression] = None
    rhs_values: List[IRExpression] = field(default_factory=list)
    is_local: bool = True
    is_loop_assignment: bool = False # Skip emission if it's just mirroring loop variables

    def __repr__(self):
        if self.rhs_call:
            return f"local {', '.join(map(str, self.targets))} = {self.rhs_call}"
        return f"local {', '.join(map(str, self.targets))} = {', '.join(map(str, self.rhs_values))}"

@dataclass(kw_only=True)
class IRBreakStatement(IRStatement):
    def __repr__(self): return "break"

@dataclass(kw_only=True)
class IRContinueStatement(IRStatement):
    def __repr__(self): return "continue"

@dataclass(kw_only=True)
class IRPassStatement(IRStatement):
    def __repr__(self): return "pass"

@dataclass(kw_only=True)
class IRComment(IRStatement):
    text: str = ""
    def __repr__(self): return f"# {self.text}"

@dataclass(kw_only=True)
class IRClassDef(IRStatement):
    name: str = ""
    bases: List[IRExpression] = field(default_factory=list)
    body: List[IRStatement] = field(default_factory=list)
    properties: Dict[str, str] = field(default_factory=dict)
    decorators: List[str] = field(default_factory=list)

@dataclass(kw_only=True)
class IRImport(IRStatement):
    module: str = ""
    alias: Optional[str] = None

@dataclass(kw_only=True)
class IRImportFrom(IRStatement):
    module: str = ""
    names: List[str] = field(default_factory=list)
    aliases: List[Optional[str]] = field(default_factory=list)
    level: int = 0

# --- CFG Terminators ---

@dataclass(kw_only=True)
class IRJump(IRStatement):
    target_block_id: int = 0
    def __repr__(self): return f"jump B{self.target_block_id}"

@dataclass(kw_only=True)
class IRBranch(IRStatement):
    condition: IRExpression = field(default_factory=lambda: IRLiteral(value=True))
    true_block_id: int = 0
    false_block_id: int = 0
    
    def __repr__(self):
        return f"branch {self.condition} then B{self.true_block_id} else B{self.false_block_id}"

@dataclass(kw_only=True)
class IRGenericForTerminator(IRStatement):
    vars: List[str] = field(default_factory=list)
    iterator: IRExpression = field(default_factory=lambda: IRVariable(name=""))
    true_block_id: int = 0  # Body
    false_block_id: int = 0 # Exit
    
    def __repr__(self):
        return f"for {', '.join(self.vars)} in {self.iterator} then B{self.true_block_id} else B{self.false_block_id}"



from typing import Optional, Any, List
from transpiler.middleend import ir, cfg, cfg_builder, ssa, purity
from transpiler.optimization import loop_detector, canonicalizer
from transpiler.backend import restructurer
from transpiler.infrastructure.flags import CompilerFlags
from transpiler.infrastructure import logger

def _optimize_block_list(stmts: List[ir.IRStatement], initial_vars: Optional[List[str]] = None, options: Optional[CompilerFlags] = None, scope_name: str = "unknown") -> List[ir.IRStatement]:
    """Helper to optimize a list of statements by building a temporary module/CFG."""
    tmp_module = ir.IRModule(body=stmts)
    
    # 1. Build CFG
    target_cfg = cfg_builder.build_cfg(tmp_module)
        
    # Validation (if enabled)
    if options and getattr(options, "verify_cfg", False):
        from transpiler.optimization.cfg_validator import CFGValidator
        validator = CFGValidator()
        validator.validate_cfg(target_cfg)
        
    # Optimization level 0: Only structural transformations required for Luau output
    opt_level = getattr(options, "opt", 2) if options else 2
    
    if opt_level == 0:
        # We still need DominatorAnalysis for Restructurer
        dom = ssa.DominatorAnalysis(target_cfg)
        dom.run()
        new_module = restructurer.restructure_cfg(target_cfg)
        canonicalizer.IRCanonicalizer().canonicalize(new_module)
        return new_module.body

    # 2. SSA Construction
    dom_analysis = ssa.DominatorAnalysis(target_cfg)
    dom_analysis.run()
    
    ssa_constructor = ssa.SSAConstructor(target_cfg, dom_analysis, initial_vars=initial_vars)
    ssa_constructor.run()
    
    # 3. Purity Analysis
    purity_analyzer = purity.PurityAnalyzer()
    for block in target_cfg.blocks.values():
        for stmt in block.statements:
            purity_analyzer.analyze_stmt(stmt)
            
    # 4. Global Optimization Fixpoint Loop
    pass_cp = ssa.SSAConstantPropagation(target_cfg)
    pass_dce = ssa.SSAMarkAndSweepDCE(target_cfg)
    pass_copy = ssa.SSACopyPropagation(target_cfg)
    pass_rte = ssa.SSARedundantTempElimination(target_cfg)
    
    # Level 1+: Basic Local Optimizations
    for i in range(5):
        changed = False
        # if pass_cp.run(): changed = True # Temporarily disabled to debug folding issues
        if pass_dce.run(initial_vars=initial_vars): changed = True
        if pass_copy.run(): changed = True
        if pass_rte.run(): changed = True
        if not changed: break
    # Level 2: Advanced Global Optimizations
    if opt_level >= 2:
        pass_cse = ssa.SSACSE(target_cfg)
        pass_cse.run()
        
        detector = loop_detector.LoopDetector(target_cfg, dom_analysis)
        detector.run()
        pass_licm = ssa.SSALICM(target_cfg, detector.loops)
        pass_licm.run()
        
        # Expression Inlining
        ssa.SSAExpressionInliner(target_cfg).run()

    # 7. SSA Destruction
    ssa.DeSSA(target_cfg).run()

    # 8. Restructuring
    dom = ssa.DominatorAnalysis(target_cfg)
    dom.run()
    new_module = restructurer.restructure_cfg(target_cfg)
        
    # 9. Canonicalization
    canonicalizer.IRCanonicalizer().canonicalize(new_module)
    
    return new_module.body

def optimize_ir(module: ir.IRModule, options: Optional[CompilerFlags] = None):
    """
    Orchestrates the RPy optimization pipeline recursively for modules and functions.
    """
    if options and options.explain:
        logger.set_explain_mode(True)
        
    # First, recursively optimize any nested functions
    def _optimize_functions(container: Any, parent_vars: List[str]):
        stmts = None
        if hasattr(container, "body") and isinstance(container.body, list): stmts = container.body
        elif isinstance(container, list): stmts = container
        
        if stmts:
            # Collect current scope variables for nested functions
            current_vars = parent_vars.copy()
            for stmt in stmts:
                if isinstance(stmt, ir.IRAssignment) and isinstance(stmt.target, ir.IRVariable):
                    current_vars.append(stmt.target.name)

            for stmt in stmts:
                if isinstance(stmt, ir.IRFunctionDef):
                    # Combine parent vars (upvalues) and function args
                    all_initial = list(set(current_vars + stmt.args))
                    # Optimize the function body BEFORE we optimize the outer scope
                    stmt.body = _optimize_block_list(stmt.body, initial_vars=all_initial, options=options, scope_name=stmt.name)
                    # Recursively check for nested functions inside this one
                    _optimize_functions(stmt.body, all_initial)

    _optimize_functions(module.body, [])

    # Finally, optimize the top-level module code
    module.body = _optimize_block_list(module.body, options=options, scope_name="module")

    # Finally, optimize the top-level module code
    module.body = _optimize_block_list(module.body, options=options, scope_name="module")
